package ru.dbotthepony.mc.otm.block

import com.google.common.collect.ImmutableMap
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import net.minecraft.ChatFormatting
import net.minecraft.Util
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.HolderLookup
import net.minecraft.core.component.DataComponents
import net.minecraft.core.particles.DustParticleOptions
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtOps
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.ComponentSerialization
import net.minecraft.util.RandomSource
import net.minecraft.world.Containers
import net.minecraft.world.InteractionResult
import net.minecraft.world.MenuProvider
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.EntityBlock
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.block.state.properties.Property
import net.minecraft.world.level.material.MapColor
import net.minecraft.world.level.material.PushReaction
import net.minecraft.world.phys.BlockHitResult
import net.minecraft.world.phys.shapes.VoxelShape
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.block.entity.IRedstoneControlled
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity.Companion
import ru.dbotthepony.mc.otm.block.entity.WorkerState
import ru.dbotthepony.mc.otm.core.TooltipList
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.asSupplier
import ru.dbotthepony.mc.otm.core.collect.SupplierMap
import ru.dbotthepony.mc.otm.core.get
import ru.dbotthepony.mc.otm.core.math.BlockRotationFreedom
import ru.dbotthepony.mc.otm.core.math.component1
import ru.dbotthepony.mc.otm.core.math.component2
import ru.dbotthepony.mc.otm.core.math.component3
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.once
import java.util.concurrent.Callable
import java.util.function.Function
import java.util.function.Supplier
import kotlin.jvm.optionals.getOrNull

private val LOGGER = LogManager.getLogger()

fun Block.getShapeForEachState(properties: List<Property<*>>, fn: (BlockState) -> VoxelShape): Map<BlockState, VoxelShape> {
	val builder = Object2ObjectArrayMap<BlockState, Supplier<VoxelShape>>()

	if (properties.isEmpty()) {
		val shape = Util.backgroundExecutor().submit(Callable { fn(stateDefinition.possibleStates.first()) }).asSupplier()

		for (state in stateDefinition.possibleStates) {
			builder[state] = shape
		}
	} else {
		val cache = Object2ObjectArrayMap<List<Any>, Supplier<VoxelShape>>()

		for (state in stateDefinition.possibleStates) {
			builder[state] = cache.computeIfAbsent(properties.map { state[it] }, Object2ObjectFunction { Util.backgroundExecutor().submit(Callable { fn(state) }).asSupplier() })
		}
	}

	return SupplierMap(builder)
}

fun Block.getShapeForEachState(property: Property<*>, fn: (BlockState) -> VoxelShape): Map<BlockState, VoxelShape> {
	return getShapeForEachState(listOf(property), fn)
}

fun interface INeighbourChangeListener {
	fun neighborChanged(
		state: BlockState,
		level: Level,
		pos: BlockPos,
		neighbour: Block,
		neighbourPos: BlockPos,
		movedByPiston: Boolean
	)
}

interface IBlockWithCustomName {
	var customDisplayName: Component?

	fun saveCustomDisplayName(nbt: CompoundTag, registry: HolderLookup.Provider) {
		if (customDisplayName != null)
			ComponentSerialization
				.CODEC
				.encodeStart(registry.createSerializationContext(NbtOps.INSTANCE), customDisplayName)
				.resultOrPartial { LOGGER.error("Unable to serialize custom name: {}", it) }
				.ifPresent { nbt["Name"] = it }
	}

	fun loadCustomDisplayName(nbt: CompoundTag, registry: HolderLookup.Provider) {
		if ("Name" in nbt) {
			customDisplayName = ComponentSerialization
				.CODEC
				.decode(registry.createSerializationContext(NbtOps.INSTANCE), nbt["Name"]!!)
				.resultOrPartial { LOGGER.error("Unable to deserialize custom name: {}", it) }
				.getOrNull()?.first
		}
	}
}

open class MatteryBlock(properties: Properties = DEFAULT_PROPERTIES) : Block(properties), INeighbourChangeListener {
	val tooltips = TooltipList()

	override fun setPlacedBy(
		level: Level,
		blockPos: BlockPos,
		blockState: BlockState,
		entity: LivingEntity?,
		itemStack: ItemStack
	) {
		if (this is EntityBlock && itemStack.has(DataComponents.CUSTOM_NAME) && !level.isClientSide) {
			val tile = level.getBlockEntity(blockPos)

			if (tile is IBlockWithCustomName) {
				try {
					tile.customDisplayName = itemStack.get(DataComponents.CUSTOM_NAME)
				} catch(_: Exception) {

				}
			}
		}

		super.setPlacedBy(level, blockPos, blockState, entity, itemStack)
	}

	@Deprecated("Use OTM specific one", level = DeprecationLevel.ERROR, replaceWith = ReplaceWith("this.getShapeForEachStateMattery"))
	override fun getShapeForEachState(mapper: Function<BlockState, VoxelShape>): ImmutableMap<BlockState, VoxelShape> {
		return super.getShapeForEachState(mapper)
	}

	fun getShapeForEachStateMattery(mapper: (BlockState) -> VoxelShape): Map<BlockState, VoxelShape> {
		return getShapeForEachState(ArrayList(stateDefinition.properties), mapper)
	}

	override fun useWithoutItem(
		blockState: BlockState,
		level: Level,
		blockPos: BlockPos,
		ply: Player,
		blockHitResult: BlockHitResult
	): InteractionResult {
		if (this is EntityBlock && !level.isClientSide) {
			val tile = level.getBlockEntity(blockPos)

			if (tile is MenuProvider) {
				ply.openMenu(tile)
				return InteractionResult.CONSUME
			}
		}

		if (this is EntityBlock && level.isClientSide)
			return InteractionResult.SUCCESS

		return super.useWithoutItem(blockState, level, blockPos, ply, blockHitResult)
	}

	override fun animateTick(blockState: BlockState, level: Level, blockPos: BlockPos, random: RandomSource) {
		if (this is EntityBlock) {
			val state = blockState.getOptionalValue(WorkerState.WORKER_STATE).or { blockState.getOptionalValue(WorkerState.SEMI_WORKER_STATE) }

			if (state.isPresent && state.get() == WorkerState.WORKING) {
				val state2 = blockState.getOptionalValue(BlockRotationFreedom.HORIZONTAL.property)
					.or { blockState.getOptionalValue(BlockRotationFreedom.DIRECTIONAL.property) }
					.or { blockState.getOptionalValue(BlockRotationFreedom.DIRECTIONAL_WITH_ROTATION.property) }

				if (state2.isPresent) {
					val direction = state2.get()

					val (x, y, z) = blockPos
					var xd = x + 0.5
					var yd = y + 0.5
					var zd = z + 0.5

					val (nx, ny, nz) = direction.normal
					xd += nx * 0.5
					yd += ny * 0.5
					zd += nz * 0.5

					when (direction.front) {
						Direction.DOWN, Direction.UP -> {
							xd += random.nextDouble() - 0.5
							zd += random.nextDouble() - 0.5
						}

						Direction.NORTH, Direction.SOUTH -> {
							xd += random.nextDouble() - 0.5
							yd += random.nextDouble() - 0.5
						}

						Direction.WEST, Direction.EAST -> {
							yd += random.nextDouble() - 0.5
							zd += random.nextDouble() - 0.5
						}
					}

					level.addParticle(DustParticleOptions.REDSTONE, xd, yd, zd, 0.0, 0.0, 0.0)
				}
			}
		}
	}

	override fun getMenuProvider(blockState: BlockState, level: Level, blockPos: BlockPos): MenuProvider? {
		if (this is EntityBlock) {
			val tile = level.getBlockEntity(blockPos)

			if (tile is MenuProvider) {
				return tile
			}
		}

		return null
	}

	override fun neighborChanged(
		state: BlockState,
		level: Level,
		pos: BlockPos,
		neighbour: Block,
		neighbourPos: BlockPos,
		movedByPiston: Boolean
	) {
		super.neighborChanged(state, level, pos, neighbour, neighbourPos, movedByPiston)

		if (this is EntityBlock && !level.isClientSide) {
			val tile = level.getBlockEntity(pos) ?: return

			level.once {
				if (!tile.isRemoved) {
					if (tile is IRedstoneControlled)
						tile.redstoneControl.redstoneSignal = level.getBestNeighborSignal(pos)

					if (tile is MatteryBlockEntity)
						tile.neighborChanged(state, level, pos, neighbour, neighbourPos, movedByPiston)
				}
			}
		}
	}

	override fun onRemove(
		oldBlockState: BlockState,
		level: Level,
		blockPos: BlockPos,
		newBlockState: BlockState,
		movedByPiston: Boolean
	) {
		if (!oldBlockState.`is`(newBlockState.block) && !level.isClientSide) {
			val blockentity = level.getBlockEntity(blockPos)

			if (blockentity is MatteryBlockEntity) {
				blockentity.beforeDroppingItems(oldBlockState, level, blockPos, newBlockState, movedByPiston)

				for (container in blockentity.droppableContainers)
					Containers.dropContents(level, blockPos, container)

				level.updateNeighbourForOutputSignal(blockPos, this)
			}
		}

		super.onRemove(oldBlockState, level, blockPos, newBlockState, movedByPiston)
	}

	override fun playerWillDestroy(level: Level, blockPos: BlockPos, blockState: BlockState, player: Player): BlockState {
		val result = super.playerWillDestroy(level, blockPos, blockState, player)

		val blockentity = level.getBlockEntity(blockPos)

		if (blockentity is MatteryBlockEntity) {
			blockentity.beforeDestroyedByPlayer(level, blockPos, blockState, player)
		}

		return result
	}

	override fun appendHoverText(
		itemStack: ItemStack,
		context: Item.TooltipContext,
		components: MutableList<Component>,
		tooltipType: TooltipFlag
	) {
		super.appendHoverText(itemStack, context, components, tooltipType)
		tooltips.assemble(itemStack, context, components)
	}

	companion object {
		val DEFAULT_PROPERTIES: Properties = Properties.of().mapColor(MapColor.METAL).requiresCorrectToolForDrops().destroyTime(1.5f).explosionResistance(25.0f)
		val DEFAULT_MACHINE_PROPERTIES: Properties = Properties.of().mapColor(MapColor.METAL).pushReaction(PushReaction.BLOCK).requiresCorrectToolForDrops().destroyTime(1.5f).explosionResistance(25.0f)
	}
}

fun <T : MatteryBlock> T.addSimpleDescription(suffix: String = "", formatting: ChatFormatting = ChatFormatting.GRAY): T {
	tooltips.add { TranslatableComponent("$descriptionId.desc$suffix").withStyle(formatting) }
	return this
}
